#include "dumpdialog.h"
#include "ui_dumpdialog.h"
#include "mainwindow.h"

#include <QMessageBox>
#include <QDomDocument>
#include <QDebug>

DumpDialog::DumpDialog(QWidget *parent, QStringList files) :
    QDialog(parent),
    ui(new Ui::DumpDialog)
{
    ui->setupUi(this);
    dirsel = new QFileDialog(this);
    dirsel->setModal(true);
    dirsel->setFileMode(QFileDialog::DirectoryOnly);

    connect(this, SIGNAL(nextSWF()), this, SLOT(processNextSWF()));
    connect(this, SIGNAL(nextImage()), this, SLOT(processNextImage()));
    connect(&active, SIGNAL(readyRead()), this, SLOT(convertData()));
    connect(&active, SIGNAL(error(QProcess::ProcessError)), this, SLOT(convertError(QProcess::ProcessError)));
    connect(&active, SIGNAL(finished(int)), this, SLOT(convertFinished(int)));
    connect(dirsel, SIGNAL(fileSelected(QString)), this, SLOT(dirSelected(QString)));
    connect(dirsel, SIGNAL(rejected()), this, SLOT(close()));

    setFixedSize(size());
    setModal(true);
    queue = files;
    setAttribute(Qt::WA_DeleteOnClose);
    show();

    total = files.size();
    current = 1;

    imgt = 0;

    ui->progressBar->setMaximum(total);

    updateContainerString();
    updateResourceString();

    dirsel->show();
}

void DumpDialog::updateContainerString(){
    ui->progressBar->setFormat(QString("File ") + QString::number(current) + " of " + QString::number(total) + " (%p%)");
}

void DumpDialog::updateResourceString(){
    if(imgt == 0)
        ui->progressBar_2->setFormat("Waiting...");
    else
        ui->progressBar_2->setFormat(QString("Resource ") + QString::number(imgc) + " of " + QString::number(imgt) + " (%p%)");
}

void DumpDialog::dirSelected(QString d){
    dir = d;
    emit nextSWF();
}

void DumpDialog::convertError(QProcess::ProcessError err){
    QString message;
    switch(err){
        case QProcess::FailedToStart:
            message = "swfmill failed to start.";
            return;
        default:
            message = "unknown error";
    }

    data.clear();
    data.squeeze();

    QMessageBox::critical(this, "Error Converting SWF", QString("While Converting ")+currentswf+":\n\t" + message);
}

void DumpDialog::convertData(){
    data.append(active.readAllStandardOutput());
}

void DumpDialog::convertFinished(int status){
    qDebug() << "swfmill exited with status" << status;
    active.close();
    if(status != 0){
        QMessageBox::critical(this, "Error Converting SWF", QString("swfmill terminated abnormally"));
        data.clear();
        data.squeeze();
        return;
    }
    QDomDocument* dom = new QDomDocument();
    if(dom->setContent(data)){
        qDebug() << "Data Parsed Successfully!";

        QDomNode swf = MainWindow::findChild(dom, "swf");
        QDomNode header = MainWindow::findChild(&swf, "Header");

        if(header.isNull()){
            qDebug() << "Couldn't Find `/swf/Header`";
        } else if(!swf.isNull()){
            QDomNode tags = MainWindow::findChild(&header, "tags");
            if(tags.isNull()) {
                qDebug() << "Couldn't Find `/swf/Header/tags`";
                goto end;
            }
            QDomNode info = MainWindow::findChild(&tags, "FrameLabel");
            if(info.isNull()){
                qDebug() << "Missing `/swf/Header/tags/FrameLabel`";
                goto end;
            }
            QDomNode symbols = MainWindow::findChild(&tags, "SymbolClass");
            if(symbols.isNull()){
                qDebug() << "Missing `/swf/Header/tags/SymbolClass`";
                goto end;
            }
            symbols = MainWindow::findChild(&symbols, "symbols");
            if(symbols.isNull()){
                qDebug() << "Missing `/swf/Header/tags/SymbolClass/symbols`";
                goto end;
            }
            QList<QDomNode> allSymbols = MainWindow::findChildren(&symbols, "Symbol");
            if(allSymbols.size() < 1){
                qDebug() << "No Symbols Found";
                goto end;
            }
            qDebug() << allSymbols.size() << "Symbols Found";

            QHash<int, QString> symbolMap;

            foreach(QDomNode node, allSymbols){
                if(!node.hasAttributes()) continue;
                if(!node.attributes().contains("objectID")) continue;
                if(!node.attributes().contains("name")) continue;
                bool ok;
                int id = node.attributes().namedItem("objectID").nodeValue().toInt(&ok);
                if(!ok) continue;
                QString name = node.attributes().namedItem("name").nodeValue();
                if(name.isEmpty() || name.isNull()) continue;
                symbolMap.insert(id, name);
            }



            QString name = MainWindow::getAttribute(&info, "label");
            if(name.isEmpty()){
                qDebug() << "FrameLabel Empty";
                goto end;
            }
            QList<QDomNode> images = MainWindow::findChildren(&tags, "DefineBitsLossless2");
            QList<QDomNode> resources = MainWindow::findChildren(&tags, "DefineBinaryData");
            qDebug() << images.size() << "Images" << resources.size() << "Resources";
            if(images.size() > 0 || resources.size() > 0){
                foreach(QDomNode node, images){
                    if(!node.hasAttributes()) continue;
                    if(!node.attributes().contains("objectID")) continue;
                    if(!node.attributes().contains("format")) continue;
                    if(!node.attributes().contains("width")) continue;
                    if(!node.attributes().contains("height")) continue;

                    int objectID, width, height, format;
                    bool ok;
                    format = node.attributes().namedItem("format").nodeValue().toInt();
                    if(format != 5) continue;
                    objectID = node.attributes().namedItem("objectID").nodeValue().toInt(&ok);
                    if(!ok) continue;
                    width = node.attributes().namedItem("width").nodeValue().toInt(&ok);
                    if(!ok) continue;
                    height = node.attributes().namedItem("height").nodeValue().toInt(&ok);
                    if(!ok) continue;

                    QString name = symbolMap.value(objectID, QString());
                    if(name.isEmpty() || name.isNull()) continue;

                    Image* img = new Image();
                    img->name = name;
                    img->w = width;
                    img->h = height;
                    img->data = node.firstChild().firstChild().firstChild().nodeValue().toLocal8Bit();
                    imgqueue << img;
                }
            } else {
                qDebug() << "No Resources In SWF";
            }
        } else {
            qDebug() << "Couldn't Find `/swf`";
        }
    } else {
        qDebug() << "Parser Error Occured";
    }

    updateContainerString();

    ui->progressBar->setValue(++current);

end:
    delete dom;
    data.clear();
    data.squeeze();
    imgt = imgqueue.size();
    if(imgt > 0){
        ui->progressBar_2->setMaximum(imgt);
        imgc = 1;
        updateResourceString();
        emit nextImage();
    } else {
        emit nextSWF();
    }
}

void DumpDialog::processNextImage(){
    Image* img = imgqueue.first();
    imgqueue.removeFirst();


    img->data = qUncompress(QByteArray("\x00\x00\xff\xff", 4) + QByteArray::fromBase64(img->data));
    if(img->data.size() > 0){
        QImage t(img->w, img->h, QImage::Format_ARGB32);
        int p=0;
        for(int y=0; y<img->h; y++){
            for(int x=0; x<img->w; x++){
                uint col = qRgba(img->data.at(p++), img->data.at(p++), img->data.at(p++), img->data.at(p++));
                col = qRgba(qBlue(col), qGreen(col), qRed(col), qAlpha(col));
                t.setPixel(x, y, col);
            }
        }
        QFile* f = new QFile(dir + QDir::separator() + img->name + ".png");
        f->open(QFile::WriteOnly);
        t.save(f, "png");
        f->close();
    }

    imgc++;

    delete img;

    if(imgqueue.size() > 0){
        ui->progressBar_2->setValue(++imgc);
        emit nextImage();
    } else {
        ui->progressBar_2->setValue(0);
        imgt = 0;
        emit nextSWF();
    }

    updateResourceString();
}

void DumpDialog::processNextSWF(){
    if(queue.size() > 0){
        currentswf = queue.first();
        queue.removeFirst();
        QStringList args;
        args << "swf2xml";
        args << currentswf;
        args << "stdout";
        qDebug() << "Converting" << currentswf;
#if defined(Q_OS_UNIX)
        active.start("swfmill", args, QProcess::ReadOnly);
#else
        active.start("swfmill.exe", args, QProcess::ReadOnly);
#endif
        return;
    }

    close();
}

DumpDialog::~DumpDialog()
{
    delete ui;
}
