#include <QDir>
#include <QMessageBox>
#include <QDebug>

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "dumpdialog.h"

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);
    dirsel = new QFileDialog(this);
    dirsel->setModal(true);
    dirsel->setFileMode(QFileDialog::DirectoryOnly);

    connect(this, SIGNAL(nextSWF()), this, SLOT(processNextSWF()));
    connect(&active, SIGNAL(readyRead()), this, SLOT(convertData()));
    connect(&active, SIGNAL(error(QProcess::ProcessError)), this, SLOT(convertError(QProcess::ProcessError)));
    connect(&active, SIGNAL(finished(int)), this, SLOT(convertFinished(int)));
    connect(ui->listWidget, SIGNAL(currentItemChanged(QListWidgetItem*,QListWidgetItem*)), this, SLOT(containerChanged(QListWidgetItem*,QListWidgetItem*)));
    connect(ui->listWidget_2, SIGNAL(currentItemChanged(QListWidgetItem*,QListWidgetItem*)), this, SLOT(resourceChanged(QListWidgetItem*,QListWidgetItem*)));
    connect(ui->cmdSetDir, SIGNAL(clicked()), this, SLOT(selectDirectory()));
    connect(ui->cmdDumpContainer, SIGNAL(clicked()), this, SLOT(dumpCurrentSWF()));
    connect(dirsel, SIGNAL(fileSelected(QString)), this, SLOT(setDirectory(QString)));

    busy = false;
    load = NULL;

    show();
    dirsel->show();
}

void MainWindow::dumpCurrentSWF(){
    QStringList files;
    files << ((SWFContainer*)ui->listWidget->currentItem()->data(Qt::UserRole).toLongLong())->filename;
    new DumpDialog(this, files);
}

void MainWindow::convertData(){
    data.append(active.readAllStandardOutput());
}

void MainWindow::updatePerc(){
    if(load == NULL) return;
    load->setText(QString::number(current) + " of "+QString::number(total)+" ("+QString::number(((float)current / (float)total) * 100.0, 'f', 2)+"%)");
}

void MainWindow::convertError(QProcess::ProcessError err){
    QString message;
    switch(err){
        case QProcess::FailedToStart:
            message = "swfmill failed to start.";
            return;
        default:
            message = "unknown error";
    }

    data.clear();
    data.squeeze();

    QMessageBox::critical(this, "Error Converting SWF", QString("While Converting ")+currentswf+":\n\t" + message);
}

QDomNode MainWindow::findChild(QDomNode* par, QString name){
    if(par->isNull()) return QDomNode();
    QDomNodeList children = par->childNodes();
    for(int i=0; i<children.size(); i++){
        if(children.at(i).nodeName() == name)
            return children.at(i);
    }
    return QDomNode();
}

QString MainWindow::getAttribute(QDomNode * node, QString attr){
    if(!node->hasAttributes() || !node->attributes().contains(attr)) return QString();
    return node->attributes().namedItem(attr).nodeValue();
}

QList<QDomNode> MainWindow::findChildren(QDomNode* par, QString name){
    QList<QDomNode> list;
    if(par->isNull()) return list;
    QDomNodeList children = par->childNodes();
    for(int i=0; i<children.size(); i++){
        if(children.at(i).nodeName() == name)
            list << children.at(i);
    }
    return list;
}


void MainWindow::convertFinished(int status){
    qDebug() << "swfmill exited with status" << status;
    active.close();
    if(status != 0){
        QMessageBox::critical(this, "Error Converting SWF", QString("swfmill terminated abnormally"));
        data.clear();
        data.squeeze();
        return;
    }
    QDomDocument* dom = new QDomDocument();
    if(dom->setContent(data)){
        qDebug() << "Data Parsed Successfully!";

        QDomNode swf = findChild(dom, "swf");
        QDomNode header = findChild(&swf, "Header");

        if(header.isNull()){
            qDebug() << "Couldn't Find `/swf/Header`";
        } else if(!swf.isNull()){
            QDomNode tags = findChild(&header, "tags");
            if(tags.isNull()) {
                qDebug() << "Couldn't Find `/swf/Header/tags`";
                goto end;
            }
            QDomNode info = findChild(&tags, "FrameLabel");
            if(info.isNull()){
                qDebug() << "Missing `/swf/Header/tags/FrameLabel`";
                goto end;
            }
            QDomNode symbols = findChild(&tags, "SymbolClass");
            if(symbols.isNull()){
                qDebug() << "Missing `/swf/Header/tags/SymbolClass`";
                goto end;
            }
            symbols = findChild(&symbols, "symbols");
            if(symbols.isNull()){
                qDebug() << "Missing `/swf/Header/tags/SymbolClass/symbols`";
                goto end;
            }
            QList<QDomNode> allSymbols = findChildren(&symbols, "Symbol");
            if(allSymbols.size() < 1){
                qDebug() << "No Symbols Found";
                goto end;
            }
            qDebug() << allSymbols.size() << "Symbols Found";

            QHash<int, QString> symbolMap;

            foreach(QDomNode node, allSymbols){
                if(!node.hasAttributes()) continue;
                if(!node.attributes().contains("objectID")) continue;
                if(!node.attributes().contains("name")) continue;
                bool ok;
                int id = node.attributes().namedItem("objectID").nodeValue().toInt(&ok);
                if(!ok) continue;
                QString name = node.attributes().namedItem("name").nodeValue();
                if(name.isEmpty() || name.isNull()) continue;
                symbolMap.insert(id, name);
            }



            QString name = getAttribute(&info, "label");
            if(name.isEmpty()){
                qDebug() << "FrameLabel Empty";
                goto end;
            }
            if(containers.contains(name)){
                qDebug() << "Already Loaded";
                goto end;
            }
            QList<QDomNode> images = findChildren(&tags, "DefineBitsLossless2");
            QList<QDomNode> resources = findChildren(&tags, "DefineBinaryData");
            qDebug() << images.size() << "Images" << resources.size() << "Resources";
            if(images.size() > 0 || resources.size() > 0){
                SWFContainer* container = new SWFContainer(name);
                container->filename = currentswf;
                containers.insert(name, container);
                ui->listWidget->addItem(container->widget);
                foreach(QDomNode node, images){
                    if(!node.hasAttributes()) continue;
                    if(!node.attributes().contains("objectID")) continue;
                    if(!node.attributes().contains("format")) continue;
                    if(!node.attributes().contains("width")) continue;
                    if(!node.attributes().contains("height")) continue;

                    int objectID, width, height, format;
                    bool ok;
                    format = node.attributes().namedItem("format").nodeValue().toInt();
                    if(format != 5) continue;
                    objectID = node.attributes().namedItem("objectID").nodeValue().toInt(&ok);
                    if(!ok) continue;
                    width = node.attributes().namedItem("width").nodeValue().toInt(&ok);
                    if(!ok) continue;
                    height = node.attributes().namedItem("height").nodeValue().toInt(&ok);
                    if(!ok) continue;

                    QString name = symbolMap.value(objectID, QString());
                    if(name.isEmpty() || name.isNull()) continue;

                    SWFResource* res = new SWFResource(container, name);
                    res->objectID = objectID;
                    res->width = width;
                    res->height = height;
                    res->type = SWFResource::BitmapImage;
                    container->resources.insert(name, res);
                }
            } else {
                qDebug() << "No Resources In SWF";
            }
        } else {
            qDebug() << "Couldn't Find `/swf`";
        }
    } else {
        qDebug() << "Parser Error Occured";
    }
end:
    delete dom;
    data.clear();
    data.squeeze();

    emit nextSWF();
}

void MainWindow::containerChanged(QListWidgetItem * itm, QListWidgetItem * old){
    SWFContainer* container;
    if(itm != NULL){
        container = (SWFContainer*)itm->data(Qt::UserRole).toLongLong();
        ui->listWidget_2->clear();
        if(container == NULL){
            ui->cmdDumpContainer->setEnabled(false);
            ui->lblContainerName->setText("No Container Selected");
        } else {
            ui->cmdDumpContainer->setEnabled(true);
            ui->lblContainerName->setText(itm->text());
            foreach(SWFResource* res, container->resources){
                ui->listWidget_2->addItem(res->getWidget());
            }
        }
    }
    if(old != NULL){
        container = (SWFContainer*)old->data(Qt::UserRole).toLongLong();
        if(container != NULL)
            container->abort();
    }
    ui->lblPrev->setText("");
    ui->lblPrev->setPixmap(NULL);
}

void MainWindow::selectDirectory(){
    dirsel->show();
}

void MainWindow::resourceChanged(QListWidgetItem * itm, QListWidgetItem *){
    if(itm == NULL) return;
    QVariant d = itm->data(Qt::UserRole);
    if(d.isNull()) return;
    SWFResource* res = (SWFResource*)d.toLongLong();
    ui->lblPrev->setText("");
    ui->lblPrev->setPixmap(NULL);
    if(res == NULL){
        ui->lblName->setText("No Resource Selected");
        ui->lblSize->setText("Unknown");
        ui->lblType->setText("Unknown");
        ui->lblWidth->setText("Unknown");
        ui->lblHeight->setText("Unknown");
    } else {
        ui->lblName->setText(res->name);
        ui->lblSize->setText("Unknown");
        ui->lblType->setText(res->typeString());
        ui->lblWidth->setText(QString::number(res->width));
        ui->lblHeight->setText(QString::number(res->height));
        res->container->fetchImage(res->objectID, ui->lblPrev);
    }
}

void MainWindow::processNextSWF(){
    if(!busy) return;
    if(queue.size() > 0){
processstart:
        current++;
        updatePerc();
        currentswf = queue.first();
        queue.removeFirst();
        if(currentswf.endsWith(".swf")){
            QStringList args;
            args << "swf2xml";
            args << currentswf;
            args << "stdout";
            qDebug() << "Converting" << currentswf;
#if defined(Q_OS_UNIX)
            active.start("swfmill", args, QProcess::ReadOnly);
#else
            active.start("swfmill.exe", args, QProcess::ReadOnly);
#endif
            return;
        }

        if(queue.size() > 0) goto processstart;
    }
    qDebug() << "Leaving";
    delete ui->listWidget->takeItem(0);
    load = NULL;
    busy = false;
    ui->cmdSetDir->setEnabled(true);
}

void MainWindow::setDirectory(QString dir){
    if(busy) return;
    busy = true;

    ui->listWidget->clear();
    ui->listWidget_2->clear();
    containers.clear();

    ui->cmdSetDir->setEnabled(false);

    qDebug() << "Entering" << dir;
    QDir d(dir);
    if(!d.exists()) return;
    QDir::setCurrent(dir);
    queue = d.entryList(QDir::Files);
    total = queue.size();
    current = -1;

    load = new QListWidgetItem(ui->listWidget);
    ui->listWidget->addItem(load);

    emit nextSWF();

    updatePerc();
}

MainWindow::~MainWindow()
{
    delete ui;
}
