#include "swfcontainer.h"
#include "swfresource.h"
#include "mainwindow.h"

#include <QDebug>
#include <QImage>

SWFContainer::SWFContainer(QString n)
{
    widget = new QListWidgetItem();
    widget->setText(n);
    widget->setData(Qt::UserRole, (qlonglong) this);
    name = n;

    active = NULL;
}

SWFContainer::~SWFContainer(){
    foreach(SWFResource* resource, resources){
        delete resource;
    }
}

void SWFContainer::convertData(){
    data.append(active->readAll());
}

void SWFContainer::convertError(QProcess::ProcessError){
    qDebug() << "swfmill Error";
    place->setText("Error Loading Sprite");
    data.clear();
    data.squeeze();

    delete active;
    active = NULL;
}

void SWFContainer::convertFinished(int code){
    if(code == 0){
        QDomDocument* dom = new QDomDocument();
        dom->setContent(data);

        QDomNode data = MainWindow::findChild(dom, "swf");
        QImage img;
        QList<QDomNode> nodes;
        QByteArray raw;
        int w, h, x, y, p;
        if(data.isNull()) goto bad;
        data = MainWindow::findChild(&data, "Header");
        if(data.isNull()) goto bad;
        data = MainWindow::findChild(&data, "tags");
        if(data.isNull()) goto bad;

        nodes = MainWindow::findChildren(&data, "DefineBitsLossless2");
        foreach(QDomNode node, nodes){
            if(!node.hasAttributes() || !node.attributes().contains("objectID")) continue;
            int objid = node.attributes().namedItem("objectID").nodeValue().toInt();
            if(objid == imgid) {
                data = node;
                break;
            }
        }
        if(!data.hasAttributes()) goto bad;
        w = data.attributes().namedItem("width").nodeValue().toInt();
        h = data.attributes().namedItem("height").nodeValue().toInt();
        if(w < 0 || h < 1) goto bad;

        data = MainWindow::findChild(&data, "data");
        data = MainWindow::findChild(&data, "data");
        if(data.isNull()) goto bad;

        raw = qUncompress(QByteArray("\x00\x00\xff\xff", 4) + QByteArray::fromBase64(data.firstChild().nodeValue().toLocal8Bit()));
        if(raw.size() < 1) goto bad;

        img = QImage(w, h, QImage::Format_ARGB32);

        p = 0;
        for(y=0; y<h; y++){
            for(x=0; x<w; x++){
                uint col = qRgba(raw.at(p++), raw.at(p++), raw.at(p++), raw.at(p++));
                col = qRgba(qBlue(col), qGreen(col), qRed(col), qAlpha(col));
                img.setPixel(x, y, col);
            }
        }

        pixmap = QPixmap::fromImage(img);

        place->setPixmap(pixmap);
        place->setText("");

        goto good;
        bad:
            place->setText("Error Extracting Data");
        good:
            delete dom;
    } else {
        place->setText("swfmill failed: " + QString::number(code));
    }

    abort();

    data.clear();
    data.squeeze();
}

void SWFContainer::abort(){
    if(active != NULL){
        disconnect(active, SIGNAL(readyRead()), this, SLOT(convertData()));
        disconnect(active, SIGNAL(error(QProcess::ProcessError)), this, SLOT(convertError(QProcess::ProcessError)));
        disconnect(active, SIGNAL(finished(int)), this, SLOT(convertFinished(int)));

        active->deleteLater();
        active = NULL;

        data.clear();
        data.squeeze();
    }

}

void SWFContainer::fetchImage(int objectID, QLabel* l){
    abort();

    active = new QProcess(this);

    connect(active, SIGNAL(readyRead()), this, SLOT(convertData()));
    connect(active, SIGNAL(error(QProcess::ProcessError)), this, SLOT(convertError(QProcess::ProcessError)));
    connect(active, SIGNAL(finished(int)), this, SLOT(convertFinished(int)));

    QStringList args;
    args << "swf2xml";
    args << filename;
    args << "stdout";
#if defined(Q_OS_UNIX)
    active->start("swfmill", args, QProcess::ReadOnly);
#else
    active->start("swfmill.exe", args, QProcess::ReadOnly);
#endif

    place = l;
    imgid = objectID;
}
